/***************************************************************/
/*                                                             */
/*  FUNCS.C                                                    */
/*                                                             */
/*  This file contains the built-in functions used in          */
/*  expressions.                                               */
/*                                                             */
/*  This file is part of REMIND.                               */
/*  Copyright (C) 1991 by David F. Skoll.                      */
/*                                                             */
/***************************************************************/
#include <stdio.h>
#include "config.h"
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include <string.h>
#include <ctype.h>
#ifdef UNIX
#ifdef HAVE_UNISTD
#include <unistd.h>
#else
#include <sys/file.h>
#endif
#endif
/* May have to change previous line to <sys/file.h> for some unixes */
#ifdef __MSDOS__
#include <io.h>
#define R_OK 4
#define W_OK 2
#define X_OK 1
#endif
#ifndef R_OK
#define R_OK 4
#define W_OK 2
#define X_OK 1
#endif
#include "types.h"
#include "globals.h"
#include "protos.h"
#include "err.h"
#include "expr.h"
#include "version.h"

/* Function prototypes */
PRIVATE	int	FAbs		ARGS ((void));
PRIVATE	int	FAccess		ARGS ((void));
PRIVATE	int	FAsc		ARGS ((void));
PRIVATE	int	FBaseyr		ARGS ((void));
PRIVATE	int	FChar		ARGS ((void));
PRIVATE	int	FChoose		ARGS ((void));
PRIVATE	int	FCoerce		ARGS ((void));
PRIVATE	int	FDate		ARGS ((void));
PRIVATE	int	FDay		ARGS ((void));
PRIVATE	int	FDaysinmon	ARGS ((void));
PRIVATE	int	FDefined	ARGS ((void));
PRIVATE	int	FFilename	ARGS ((void));
PRIVATE	int	FGetenv		ARGS ((void));
PRIVATE	int	FHour		ARGS ((void));
PRIVATE	int	FIif		ARGS ((void));
PRIVATE	int	FIndex		ARGS ((void));
PRIVATE	int	FIsomitted	ARGS ((void));
PRIVATE	int	FMax		ARGS ((void));
PRIVATE	int	FMin		ARGS ((void));
PRIVATE	int	FMinute		ARGS ((void));
PRIVATE	int	FMon		ARGS ((void));
PRIVATE	int	FMonnum		ARGS ((void));
PRIVATE	int	FOrd		ARGS ((void));
PRIVATE	int	FOstype 	ARGS ((void));
PRIVATE	int	FPlural		ARGS ((void));
PRIVATE	int	FSgn		ARGS ((void));
PRIVATE	int	FShell		ARGS ((void));
PRIVATE	int	FStrlen		ARGS ((void));
PRIVATE	int	FSubstr		ARGS ((void));
PRIVATE	int	FTime		ARGS ((void));
PRIVATE	int	FTrigdate	ARGS ((void));
PRIVATE	int	FTrigtime	ARGS ((void));
PRIVATE	int	FTrigvalid	ARGS ((void));
PRIVATE	int	FTypeof		ARGS ((void));
PRIVATE	int	FUpper		ARGS ((void));
PRIVATE	int	FValue		ARGS ((void));
PRIVATE	int	FVersion	ARGS ((void));
PRIVATE	int	FWkday		ARGS ((void));
PRIVATE	int	FWkdaynum	ARGS ((void));
PRIVATE	int	FYear		ARGS ((void));
PRIVATE int	FIsleap         ARGS ((void));
PRIVATE int	FLower          ARGS ((void));
PRIVATE int	FNow            ARGS ((void));
PRIVATE int	FRealtoday      ARGS ((void));
PRIVATE int	FToday          ARGS ((void));
PRIVATE int	FTrigger        ARGS ((void));
PRIVATE int	CheckArgs       ARGS ((Operator *f, int nargs));
PRIVATE int	CleanUpAfterFunc ARGS ((void));
     
#ifdef __MSDOS__
PRIVATE FILE *popen  ARGS((char *cmd, char *mode));
PRIVATE int   pclose ARGS((FILE *fp));
#endif

/* "Overload" the struct Operator definition */
#define NO_MAX 127
#define MINARGS prec
#define MAXARGS type

/* Sigh - we use a global var. to hold the number of args supplied to
   function being called */
static int Nargs;

/* Use a global var. to hold function return value */
static Value RetVal;

/* Temp string buffer */
static char Buffer[32];

/* We need access to the value stack */
extern Value ValStack[];
extern int ValStackPtr;

/* Macro for accessing arguments from the value stack - args are numbered
   from 0 to (Nargs - 1) */
#define ARG(x) (ValStack[ValStackPtr - Nargs + (x)])

/* Macro for copying a value while destroying original copy */
#define DCOPYVAL(x, y) ( (x) = (y), (y).type = ERR_TYPE )

/* Convenience macros */
#define UPPER(c) ( ((c) >= 'a' && (c) <= 'z') ? (c) - 'a' + 'A' : (c) )
#define LOWER(c) ( ((c) >= 'A' && (c) <= 'Z') ? (c) + 'a' - 'A' : (c) )

/* The array holding the built-in functions. */
Operator Func[] = {
/*	Name		minargs maxargs	func   */

    {   "abs",		1,	1,	FAbs	},
    {   "access",       2,      2,      FAccess },
    {   "asc",		1,	1,	FAsc	},
    {   "baseyr",	0,	0,	FBaseyr	},
    {   "char",		1,	NO_MAX,	FChar	},
    {   "choose",	2,	NO_MAX, FChoose },
    {   "coerce",	2,	2,	FCoerce },
    {   "date",		3,	3,	FDate	},
    {   "day",		1,	1,	FDay	},
    {   "daysinmon",	2,	2,	FDaysinmon },
    {   "defined",	1,	1,	FDefined },
    {   "filename",	0,	0,	FFilename },
    {   "getenv",	1,	1,	FGetenv },
    {   "hour",		1,	1,	FHour	},
    {   "iif",		3,	3,	FIif	},
    {   "index",	2,	3,	FIndex 	},
    {   "isleap",	1,	1,	FIsleap },
    {   "isomitted",	1,	1,	FIsomitted },
    {   "lower",	1,	1,	FLower	},
    {   "max",		1,	NO_MAX,	FMax	},
    {   "min",		1,	NO_MAX, FMin	},
    {   "minute",	1,	1,	FMinute },
    {   "mon",		1,	1,	FMon	},
    {   "monnum",	1,	1,	FMonnum },
    {   "now",		0,	0,	FNow	},
    {   "ord",		1,	1,	FOrd	},
    {   "ostype",       0,      0,      FOstype },
    {   "plural",	1,	3,	FPlural },
    {   "realtoday",    0,      0,      FRealtoday },
    {   "sgn",		1,	1,	FSgn	},
    {   "shell",	1,	1,	FShell	},
    {   "strlen",	1,	1,	FStrlen	},
    {   "substr",	2,	3,	FSubstr	},
    {   "time",		2,	2,	FTime	},
    {   "today",	0,	0,	FToday	},
    {   "trigdate",	0,	0,	FTrigdate },
    {   "trigger",	1,	1,	FTrigger },
    {   "trigtime",	0,	0,	FTrigtime },
    {   "trigvalid",	0,	0,	FTrigvalid },
    {   "typeof",       1,      1,      FTypeof },
    {   "upper",	1,	1,	FUpper	},
    {   "value",	1,	2,	FValue	},
    {   "version",      0,      0,      FVersion },
    {   "wkday",	1,	1,	FWkday	},
    {   "wkdaynum",	1,	1,	FWkdaynum },
    {   "year",		1,	1,	FYear	}
};

/* Need a variable here - Func[] array not really visible to outside. */
int NumFuncs = sizeof(Func) / sizeof(Operator) ;
     
/***************************************************************/
/*                                                             */
/*  CallFunc                                                   */
/*                                                             */
/*  Call a function given a pointer to it, and the number      */
/*  of arguments supplied.                                     */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PUBLIC int CallFunc(Operator *f, int nargs)
#else
int CallFunc(f, nargs)
Operator *f;
int nargs;
#endif
{
   register int r = CheckArgs(f, nargs);
   int i;

   Nargs = nargs;

   RetVal.type = ERR_TYPE;
   if (DebugFlag & DB_PRTEXPR) {
      fprintf(ErrFp, "%s(", f->name);
      for (i=0; i<nargs; i++) {
	 PrintValue(&ARG(i), ErrFp);
	 if (i<nargs-1) fprintf(ErrFp, ", ");
      }
      fprintf(ErrFp, ") => ");
      if (r) {
	 fprintf(ErrFp, "%s\n", ErrMsg[r]);
	 return r;
      }
   }
   if (r) {
      Eprint("%s(): %s", f->name, ErrMsg[r]);
      return r;
   }

   r = (*(f->func))();
   if (r) {
      DestroyValue(&RetVal);
      if (DebugFlag & DB_PRTEXPR)
	 fprintf(ErrFp, "%s\n", ErrMsg[r]);
      else
	 Eprint("%s(): %s", f->name, ErrMsg[r]);
      return r;
   }
   if (DebugFlag & DB_PRTEXPR) {
      PrintValue(&RetVal, ErrFp);
      fprintf(ErrFp, "\n");
   }
   r = CleanUpAfterFunc();
   return r;
}

/***************************************************************/
/*                                                             */
/*  CheckArgs                                                  */
/*                                                             */
/*  Check that the right number of args have been supplied     */
/*  for a function.                                            */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int CheckArgs(Operator *f, int nargs)
#else
static int CheckArgs(f, nargs)
Operator *f;
int nargs;
#endif
{
   if (nargs < f->MINARGS) return E_2FEW_ARGS;
   if (nargs > f->MAXARGS && f->MAXARGS != NO_MAX) return E_2MANY_ARGS;
   return OK;
}
/***************************************************************/
/*                                                             */
/*  CleanUpAfterFunc                                           */
/*                                                             */
/*  Clean up the stack after a function call - remove          */
/*  args and push the new value.                               */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int CleanUpAfterFunc(void)
#else
static int CleanUpAfterFunc()
#endif
{
   Value v;
   int i, r;

   for (i=0; i<Nargs; i++) {
      r = PopValStack(&v);
      if (r) return r;
      DestroyValue(&v);
   }
   PushValStack(&RetVal);
   return OK;
}

/***************************************************************/
/*                                                             */
/*  RetStrVal                                                  */
/*                                                             */
/*  Return a string value from a function.                     */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int RetStrVal(const char *s)
#else
static int RetStrVal(s)
char *s;
#endif
{
   RetVal.type = STR_TYPE;
   if (!s) {
      RetVal.v.str = (char *) malloc(1);
      if (RetVal.v.str) *RetVal.v.str = 0;
   } else
      RetVal.v.str = StrDup(s);

   if (!RetVal.v.str) {
      RetVal.type = ERR_TYPE;
      return E_NO_MEM;
   }
   return OK;
}


/***************************************************************/
/*                                                             */
/*  FStrlen - string length                                    */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FStrlen(void)
#else
static int FStrlen()
#endif
{
   Value *v = &ARG(0);
   if (v->type != STR_TYPE) return E_BAD_TYPE;
   RetVal.type = INT_TYPE;
   RetVal.v.val = strlen(v->v.str);
   return OK;
}

/***************************************************************/
/*                                                             */
/*  FBaseyr - system base year                                 */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FBaseyr(void)
#else
static int FBaseyr()
#endif
{
   RetVal.type = INT_TYPE;
   RetVal.v.val = BASE;
   return OK;
}

/***************************************************************/
/*                                                             */
/*  FDate - make a date from year, month, day.                 */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FDate(void)
#else
static int FDate()
#endif
{
   int y, m, d;
   if (ARG(0).type != INT_TYPE ||
       ARG(1).type != INT_TYPE ||
       ARG(2).type != INT_TYPE) return E_BAD_TYPE;
   y = ARG(0).v.val;
   m = ARG(1).v.val - 1;
   d = ARG(2).v.val;

   if (y < BASE || y > BASE+YR_RANGE || m < 0 || m > 11 || d < 1 ||
       d > DaysInMonth(m, y)) return E_BAD_DATE;

   RetVal.type = DATE_TYPE;
   RetVal.v.val = Julian(y, m, d);
   return OK;
}

/***************************************************************/
/*                                                             */
/*  FCoerce - type coercion function.                          */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FCoerce(void)
#else
static int FCoerce()
#endif
{
   char *s;

   if (ARG(0).type != STR_TYPE) return E_BAD_TYPE;
   s = ARG(0).v.str;

   /* Copy the value of ARG(1) into RetVal, and make ARG(1) invalid so
      it won't be destroyed */
   DCOPYVAL(RetVal, ARG(1));

   if (StriEq(s, "int")) return DoCoerce(INT_TYPE, &RetVal);
   else if (StriEq(s, "date")) return DoCoerce(DATE_TYPE, &RetVal);
   else if (StriEq(s, "time")) return DoCoerce(TIM_TYPE, &RetVal);
   else if (StriEq(s, "string")) return DoCoerce(STR_TYPE, &RetVal);
   else return E_CANT_COERCE;
}

/***************************************************************/
/*                                                             */
/*  FMax - select maximum from a list of args.                 */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FMax(void)
#else
static int FMax()
#endif
{
   Value *maxptr;
   int i;
   char type;

   maxptr = &ARG(0);
   type = maxptr->type;

   for (i=1; i<Nargs; i++) {
      if (ARG(i).type != type) return E_BAD_TYPE;
      if (type != STR_TYPE) {
	 if (ARG(i).v.val > maxptr->v.val) maxptr=&ARG(i);
      } else {
	 if (strcmp(ARG(i).v.str, maxptr->v.str) > 0) maxptr=&ARG(i);
      }
   }
   DCOPYVAL(RetVal, *maxptr);
   return OK;
}

/***************************************************************/
/*                                                             */
/*  FMin - select minimum from a list of args.                 */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FMin(void)
#else
static int FMin()
#endif
{
   Value *minptr;
   int i;
   char type;

   minptr = &ARG(0);
   type = minptr->type;

   for (i=1; i<Nargs; i++) {
      if (ARG(i).type != type) return E_BAD_TYPE;
      if (type != STR_TYPE) {
	 if (ARG(i).v.val < minptr->v.val) minptr=&ARG(i);
      } else {
	 if (strcmp(ARG(i).v.str, minptr->v.str) < 0) minptr=&ARG(i);
      }
   }
   DCOPYVAL(RetVal, *minptr);
   return OK;
}

/***************************************************************/
/*                                                             */
/*  FAsc - ASCII value of first char of string                 */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FAsc(void)
#else
static int FAsc()
#endif
{
   if (ARG(0).type != STR_TYPE) return E_BAD_TYPE;
   RetVal.type = INT_TYPE;
   RetVal.v.val = *(ARG(0).v.str);
   return OK;
}

/***************************************************************/
/*                                                             */
/*  FChar - build a string from ASCII values                   */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FChar(void)
#else
static int FChar()
#endif
{

   int i, len;

/* Special case of one arg - if given ascii value 0, create empty string */
   if (Nargs == 1) {
      if (ARG(0).type != INT_TYPE) return E_BAD_TYPE;
      if (ARG(0).v.val < 0) return E_2LOW;
      if (ARG(0).v.val > 255) return E_2HIGH;
      len = ARG(0).v.val ? 2 : 1;
      RetVal.v.str = (char *) malloc(len);
      if (!RetVal.v.str) return E_NO_MEM;
      RetVal.type = STR_TYPE;
      *(RetVal.v.str) = ARG(0).v.val;
      if (len>1) *(RetVal.v.str + 1) = 0;
      return OK;
   }

   RetVal.v.str = (char *) malloc(Nargs + 1);
   if (!RetVal.v.str) return E_NO_MEM;
   RetVal.type = STR_TYPE;
   for (i=0; i<Nargs; i++) {
     if (ARG(i).type != INT_TYPE) return E_BAD_TYPE;
     if (ARG(i).v.val < 1) return E_2LOW;
     if (ARG(i).v.val > 255) return E_2HIGH;
     *(RetVal.v.str + i) = ARG(i).v.val;
   }
   *(RetVal.v.str + Nargs) = 0;
   return OK;
}
/***************************************************************/
/*                                                             */
/*  Functions for extracting the components of a date.         */
/*                                                             */
/*  FDay - get day of month                                    */
/*  FMonnum - get month (1-12)                                 */
/*  FYear - get year                                           */
/*  FWkdaynum - get weekday num (0 = Sun)                      */
/*  FWkday - get weekday (string)                              */
/*  FMon - get month (string)                                  */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FDay(void)
#else
static int FDay()
#endif
{
   int y, m, d;
   if (ARG(0).type != DATE_TYPE) return E_BAD_TYPE;
   FromJulian(ARG(0).v.val, &y, &m, &d);
   RetVal.type = INT_TYPE;
   RetVal.v.val = d;
   return OK;
}

#ifdef HAVE_PROTOS
PRIVATE int FMonnum(void)
#else
static int FMonnum()
#endif
{
   int y, m, d;
   if (ARG(0).type != DATE_TYPE) return E_BAD_TYPE;
   FromJulian(ARG(0).v.val, &y, &m, &d);
   RetVal.type = INT_TYPE;
   RetVal.v.val = m+1;
   return OK;
}

#ifdef HAVE_PROTOS
PRIVATE int FYear(void)
#else
static int FYear()
#endif
{
   int y, m, d;
   if (ARG(0).type != DATE_TYPE) return E_BAD_TYPE;
   FromJulian(ARG(0).v.val, &y, &m, &d);
   RetVal.type = INT_TYPE;
   RetVal.v.val = y;
   return OK;
}

#ifdef HAVE_PROTOS
PRIVATE int FWkdaynum(void)
#else
static int FWkdaynum()
#endif
{
   if (ARG(0).type != DATE_TYPE) return E_BAD_TYPE;
   RetVal.type = INT_TYPE;

   /* Correct so that 0 = Sunday */
   RetVal.v.val = (ARG(0).v.val+1) % 7;
   return OK;
}

#ifdef HAVE_PROTOS
PRIVATE int FWkday(void)
#else
static int FWkday()
#endif
{
   char *s;

   if (ARG(0).type != DATE_TYPE && ARG(0).type != INT_TYPE) return E_BAD_TYPE;
   if (ARG(0).type == INT_TYPE) {
      if (ARG(0).v.val < 0) return E_2LOW;
      if (ARG(0).v.val > 6) return E_2HIGH;
      /* Convert 0=Sun to 0=Mon */
      ARG(0).v.val--;
      if (ARG(0).v.val < 0) ARG(0).v.val = 6;
      s = DayName[ARG(0).v.val];
   } else s = DayName[ARG(0).v.val % 7];
   return RetStrVal(s);
}

#ifdef HAVE_PROTOS
PRIVATE int FMon(void)
#else
static int FMon()
#endif
{
   char *s;
   int y, m, d;

   if (ARG(0).type != DATE_TYPE && ARG(0).type != INT_TYPE)
      return E_BAD_TYPE;
   if (ARG(0).type == INT_TYPE) {
      m = ARG(0).v.val - 1;
      if (m < 0) return E_2LOW;
      if (m > 11) return E_2HIGH;
   } else FromJulian(ARG(0).v.val, &y, &m, &d);
   s = MonthName[m];
   return RetStrVal(s);
}

/***************************************************************/
/*                                                             */
/*  FHour - extract hour from a time                           */
/*  FMinute - extract minute from a time                       */
/*  FTime - create a time from hour and minute                 */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FHour(void)
#else
static int FHour()
#endif
{
   if (ARG(0).type != TIM_TYPE) return E_BAD_TYPE;
   RetVal.type = INT_TYPE;
   RetVal.v.val = ARG(0).v.val / 60;
   return OK;
}

#ifdef HAVE_PROTOS
PRIVATE int FMinute(void)
#else
static int FMinute()
#endif
{
   if (ARG(0).type != TIM_TYPE) return E_BAD_TYPE;
   RetVal.type = INT_TYPE;
   RetVal.v.val = ARG(0).v.val % 60;
   return OK;
}

#ifdef HAVE_PROTOS
PRIVATE int FTime(void)
#else
static int FTime()
#endif
{
   int h, m;

   if (ARG(0).type != INT_TYPE || ARG(1).type != INT_TYPE) return E_BAD_TYPE;

   h = ARG(0).v.val;
   m = ARG(1).v.val;
   if (h<0 || m<0) return E_2LOW;
   if (h>23 || m>59) return E_2HIGH;
   RetVal.type = TIM_TYPE;
   RetVal.v.val = h*60 + m;
   return OK;
}

/***************************************************************/
/*                                                             */
/*  FAbs - absolute value                                      */
/*  FSgn - signum function                                     */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FAbs(void)
#else
static int FAbs()
#endif
{
   int v;

   if (ARG(0).type != INT_TYPE) return E_BAD_TYPE;
   v = ARG(0).v.val;
   RetVal.type = INT_TYPE;
   RetVal.v.val = (v < 0) ? (-v) : v;
   return OK;
}

#ifdef HAVE_PROTOS
PRIVATE int FSgn(void)
#else
static int FSgn()
#endif
{
   int v;

   if (ARG(0).type != INT_TYPE) return E_BAD_TYPE;
   v = ARG(0).v.val;
   RetVal.type = INT_TYPE;
   if (v>0) RetVal.v.val = 1;
   else if (v<0) RetVal.v.val = -1;
   else RetVal.v.val = 0;
   return OK;
}

/***************************************************************/
/*                                                             */
/*  FOrd - returns a string containing ordinal number.         */
/*                                                             */
/*  EG - ord(2) == "2nd", etc.                                 */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FOrd(void)
#else
static int FOrd()
#endif
{
   int t, u, v;
   char *s;

   if (ARG(0).type != INT_TYPE) return E_BAD_TYPE;

   v = ARG(0).v.val;
   t = v % 100;
   if (t < 0) t = -t;
   u = t % 10;
   s = "th";
   if (u == 1 && t != 11) s = "st";
   if (u == 2 && t != 12) s = "nd";
   if (u == 3 && t != 13) s = "rd";
   sprintf(Buffer, "%d%s", v, s);
   return RetStrVal(Buffer);
}

/***************************************************************/
/*                                                             */
/*  FPlural - pluralization function                           */
/*                                                             */
/*  plural(n) -->  "" or "s"                                   */
/*  plural(n, str) --> "str" or "strs"                         */
/*  plural(n, str1, str2) --> "str1" or "str2"                 */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FPlural(void)
#else
static int FPlural()
#endif
{
   if (ARG(0).type != INT_TYPE) return E_BAD_TYPE;

   switch(Nargs) {
      case 1:
	 if (ARG(0).v.val == 1) return RetStrVal("");
	 else return RetStrVal("s");

      case 2:
	 if (ARG(1).type != STR_TYPE) return E_BAD_TYPE;
	 if (ARG(0).v.val == 1) {
	    DCOPYVAL(RetVal, ARG(1));
	    return OK;
	 }
	 RetVal.type = STR_TYPE;
	 RetVal.v.str = (char *) malloc(strlen(ARG(1).v.str)+2);
	 if (!RetVal.v.str) {
	    RetVal.type = ERR_TYPE;
	    return E_NO_MEM;
	 }
	 strcpy(RetVal.v.str, ARG(1).v.str);
	 strcat(RetVal.v.str, "s");
	 return OK;

      default:
	 if (ARG(1).type != STR_TYPE || ARG(2).type != STR_TYPE)
	    return E_BAD_TYPE;
	 if (ARG(0).v.val == 1) DCOPYVAL(RetVal, ARG(1));
	 else DCOPYVAL(RetVal, ARG(2));
	 return OK;
   }
}

/***************************************************************/
/*                                                             */
/*  FChoose                                                    */
/*  Choose the nth value from a list of value.  If n<1, choose */
/*  first.  If n>N, choose Nth value.  Indexes always start    */
/*  from 1.                                                    */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FChoose(void)
#else
static int FChoose()
#endif
{
   int v;

   if (ARG(0).type != INT_TYPE) return E_BAD_TYPE;
   v = ARG(0).v.val;
   if (v < 1) v = 1;
   if (v > Nargs-1) v = Nargs-1;
   DCOPYVAL(RetVal, ARG(v));
   return OK;
}

/***************************************************************/
/*                                                             */
/*  FVersion - version of Remind                               */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FVersion(void)
#else
static int FVersion()
#endif
{
   return RetStrVal(VERSION);
}

/***************************************************************/
/*                                                             */
/*  FOstype - the type of operating system (UNIX or MSDOS)     */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FOstype(void)
#else
static int FOstype()
#endif
{
#ifdef UNIX
   return RetStrVal("UNIX");
#else
   return RetStrVal("MSDOS");
#endif  
}

/***************************************************************/
/*                                                             */
/*  FUpper - convert string to upper-case                      */
/*  FLower - convert string to lower-case                      */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FUpper(void)
#else
static int FUpper()
#endif
{
   char *s;

   if (ARG(0).type != STR_TYPE) return E_BAD_TYPE;
   DCOPYVAL(RetVal, ARG(0));
   s = RetVal.v.str;
   while (*s) { *s = UPPER(*s); s++; }
   return OK;
}

#ifdef HAVE_PROTOS
PRIVATE int FLower(void)
#else
static int FLower()
#endif
{
   char *s;

   if (ARG(0).type != STR_TYPE) return E_BAD_TYPE;
   DCOPYVAL(RetVal, ARG(0));
   s = RetVal.v.str;
   while (*s) { *s = LOWER(*s); s++; }
   return OK;
}

/***************************************************************/
/*                                                             */
/*  FToday - return the system's notion of "today"             */
/*  Frealtoday - return today's date as read from OS.          */
/*  FNow - return the system time                              */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FToday(void)
#else
static int FToday()
#endif
{
   RetVal.type = DATE_TYPE;
   RetVal.v.val = JulianToday;
   return OK;
}

#ifdef HAVE_PROTOS
PRIVATE int FRealtoday(void)
#else
static int FRealtoday()
#endif
{
   RetVal.type = DATE_TYPE;
   RetVal.v.val = RealToday;
   return OK;
}

#ifdef HAVE_PROTOS
PRIVATE int FNow(void)
#else
static int FNow()
#endif
{
   RetVal.type = TIM_TYPE;
   RetVal.v.val = (int) ( SystemTime() / 60L );
   return OK;
}

/***************************************************************/
/*                                                             */
/*  FGetenv - get the value of an environment variable.        */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FGetenv(void)
#else
static int FGetenv()
#endif
{
   if (ARG(0).type != STR_TYPE) return E_BAD_TYPE;
   return RetStrVal(getenv(ARG(0).v.str));
}
     
/***************************************************************/
/*                                                             */
/*  FValue                                                     */
/*                                                             */
/*  Get the value of a variable.  If a second arg is supplied, */
/*  it is returned if variable is undefined.                   */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FValue(void)
#else
static int FValue()
#endif
{
   Var *v;

   if (ARG(0).type != STR_TYPE) return E_BAD_TYPE;
   switch(Nargs) {
      case 1:
	 return GetVarValue(ARG(0).v.str, &RetVal, NULL);

      case 2:
         v = FindVar(ARG(0).v.str, 0);
	 if (!v) {
	    DCOPYVAL(RetVal, ARG(1));
	    return OK;
	 } else {
	    return CopyValue(&RetVal, &v->v);
	 }
   }
   return OK;
}

/***************************************************************/
/*                                                             */
/*  FDefined                                                   */
/*                                                             */
/*  Return 1 if a variable is defined, 0 if it is not.         */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FDefined(void)
#else
static int FDefined()
#endif
{
   if (ARG(0).type != STR_TYPE) return E_BAD_TYPE;

   RetVal.type = INT_TYPE;

   if (FindVar(ARG(0).v.str, 0))
      RetVal.v.val = 1;
   else
      RetVal.v.val = 0;
   return OK;
}

/***************************************************************/
/*                                                             */
/*  FTrigdate and FTrigtime                                    */
/*                                                             */
/*  Date and time of last trigger.  These are stored in global */
/*  vars.                                                      */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FTrigdate(void)
#else
static int FTrigdate()
#endif
{
   RetVal.type = DATE_TYPE;
   RetVal.v.val = LastTriggerDate;
   return OK;
}

#ifdef HAVE_PROTOS
PRIVATE int FTrigvalid(void)
#else
static int FTrigvalid()
#endif
{
   RetVal.type = INT_TYPE;
   RetVal.v.val = LastTrigValid;
   return OK;
}

#ifdef HAVE_PROTOS
PRIVATE int FTrigtime(void)
#else
static int FTrigtime()
#endif
{
   RetVal.type = TIM_TYPE;
   RetVal.v.val = LastTriggerTime;
   return OK;
}

/***************************************************************/
/*                                                             */
/*  FDaysinmon                                                 */
/*                                                             */
/*  Returns the number of days in mon,yr                       */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FDaysinmon(void)
#else
static int FDaysinmon()
#endif
{
   if (ARG(0).type != INT_TYPE || ARG(1).type != INT_TYPE) return E_BAD_TYPE;

   if (ARG(0).v.val > 12 || ARG(0).v.val < 1 ||
       ARG(1).v.val < BASE || ARG(1).v.val > BASE+YR_RANGE)
          return E_DOMAIN_ERR;

   RetVal.type = INT_TYPE;
   RetVal.v.val = DaysInMonth(ARG(0).v.val-1, ARG(1).v.val);
   return OK;
}

/***************************************************************/
/*                                                             */
/*  FIsleap                                                    */
/*                                                             */
/*  Return 1 if year is a leap year, zero otherwise.           */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FIsleap(void)
#else
static int FIsleap()
#endif
{
   int y, m, d;

   if (ARG(0).type != INT_TYPE && ARG(0).type != DATE_TYPE) return E_BAD_TYPE;

   /* If it's a date, extract the year */
   if (ARG(0).type == DATE_TYPE)
      FromJulian(ARG(0).v.val, &y, &m, &d);
   else
      y = ARG(0).v.val;

   RetVal.type = INT_TYPE;
   RetVal.v.val = IsLeapYear(y);
   return OK;
}

/***************************************************************/
/*                                                             */
/*  FTrigger                                                   */
/*                                                             */
/*  Put out a date in a format suitable for triggering.        */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FTrigger(void)
#else
static int FTrigger()
#endif
{
   int y, m, d;
   char buf[40];

   if (ARG(0).type != DATE_TYPE) return E_BAD_TYPE;

   FromJulian(ARG(0).v.val, &y, &m, &d);
   sprintf(buf, "%d %s %d", d, MonthName[m], y);
   return RetStrVal(buf);
}

/***************************************************************/
/*                                                             */
/*  FShell                                                     */
/*                                                             */
/*  The shell function.                                        */
/*                                                             */
/*  If run is disabled, will not be executed.                  */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FShell(void)
#else
static int FShell()
#endif
{
   char buf[SHELLSIZE+1];
   int ch, len;
   FILE *fp;
   char *s;

   if (RunDisabled) return E_RUN_DISABLED;
   if (ARG(0).type != STR_TYPE) return E_BAD_TYPE;
   s = buf;
   len = 0;
   fp = popen(ARG(0).v.str, "r");
   if (!fp) return E_IO_ERR;
   while (len < SHELLSIZE) {
      ch = getc(fp);
      if (ch == EOF) {
	 break;
      }
      if (isspace(ch)) *s++ = ' ';
      else            *s++ = ch;
      len++;
   }
   *s = 0;

   /* Delete trailing newline (converted to space) */
   if (s > buf && *(s-1) == ' ') *(s-1) = 0;
#ifdef __MSDOS__
   if (s-1 > buf && *(s-2) == ' ') *(s-2) = 0;
#endif
   pclose(fp);
   return RetStrVal(buf);
}

/***************************************************************/
/*                                                             */
/*  FIsomitted                                                 */
/*                                                             */
/*  Is a date omitted?                                         */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FIsomitted(void)
#else
static int FIsomitted()
#endif
{
   if (ARG(0).type != DATE_TYPE) return E_BAD_TYPE;
   RetVal.type = INT_TYPE;
   RetVal.v.val = IsOmitted(ARG(0).v.val, 0);
   return OK;
}

/***************************************************************/
/*                                                             */
/*  FSubstr                                                    */
/*                                                             */
/*  The substr function.  We destroy the value on the stack.   */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FSubstr(void)
#else
static int FSubstr()
#endif
{
   char *s, *t;
   int start, end;

   if (ARG(0).type != STR_TYPE || ARG(1).type != INT_TYPE) return E_BAD_TYPE;
   if (Nargs == 3 && ARG(2).type != INT_TYPE) return E_BAD_TYPE;

   s = ARG(0).v.str;
   start = 1;
   while (start < ARG(1).v.val) {
      if (!*s) break;
      s++;
      start++;
   }
   if (Nargs == 2 || !*s) return RetStrVal(s);
   end = start;
   t = s;
   while (end <= ARG(2).v.val) {
      if (!*s) break;
      s++;
      end++;
   }
   *s = 0;
   return RetStrVal(t);
}

/***************************************************************/
/*                                                             */
/*  FIndex                                                     */
/*                                                             */
/*  The index of one string embedded in another.               */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FIndex(void)
#else
static int FIndex()
#endif
{
   char *s;
   int start;

   if (ARG(0).type != STR_TYPE || ARG(1).type != STR_TYPE ||
       (Nargs == 3 && ARG(2).type != INT_TYPE)) return E_BAD_TYPE;

   s = ARG(0).v.str;

/* If 3 args, bump up the start */
   if (Nargs == 3) {
      start = 1;
      while (start < ARG(2).v.val) {
         if (!*s) break;
	 s++;
	 start++;
      }
  }

/* Find the string */
   s = strstr(s, ARG(1).v.str);
   RetVal.type = INT_TYPE;
   if (!s) {
      RetVal.v.val = 0;
      return OK;
   }
   RetVal.v.val = (s - ARG(0).v.str) + 1;
   return OK;
}

/***************************************************************/
/*                                                             */
/*  FIif                                                       */
/*                                                             */
/*  The IIF function.                                          */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FIif(void)
#else
static int FIif()
#endif
{
   int istrue;

   if (ARG(0).type != STR_TYPE && ARG(0).type != INT_TYPE) return E_BAD_TYPE;

   if (ARG(0).type == INT_TYPE)
      istrue = ARG(0).v.val;
   else
      istrue = *(ARG(0).v.str);

   if (istrue) DCOPYVAL(RetVal, ARG(1));
   else        DCOPYVAL(RetVal, ARG(2));

   return OK;
}

/***************************************************************/
/*                                                             */
/*  FFilename                                                  */
/*                                                             */
/*  Return name of current file                                */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FFilename(void)
#else
static int FFilename()
#endif
{
   return RetStrVal(FileName);
}

/***************************************************************/
/*                                                             */
/*  FAccess                                                    */
/*                                                             */
/*  The UNIX access() system call.                             */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FAccess(void)
#else
static int FAccess()
#endif
{
   int amode;
   char *s;
  
   if (ARG(0).type != STR_TYPE || 
       (ARG(1).type != INT_TYPE && ARG(1).type != STR_TYPE)) return E_BAD_TYPE;

   if (ARG(1).type == INT_TYPE) amode = ARG(1).v.val;
   else {
      amode = 0;
      s = ARG(1).v.str;
      while (*s) {
         switch(*s++) {
	    case 'r':
	    case 'R': amode |= R_OK; break;
	    case 'w':
	    case 'W': amode |= W_OK; break;
	    case 'x':
	    case 'X': amode |= X_OK; break;
         }
      }
   } 
   RetVal.type = INT_TYPE;
   RetVal.v.val = access(ARG(0).v.str, amode);
   return OK;
}

#ifdef __MSDOS__
/***************************************************************/
/*                                                             */
/*  popen and pclose                                           */
/*                                                             */
/*  These are some rather brain-dead kludges for MSDOS.        */
/*  They are just sufficient for the shell() function, and     */
/*  should NOT be viewed as general-purpose replacements       */
/*  for the UNIX system calls.                                 */
/*                                                             */
/***************************************************************/
#ifdef __TURBOC__
#pragma argsused
#endif

static char *TmpFile;
#ifdef HAVE_PROTOS
PRIVATE FILE *popen(char *cmd, char *mode)
#else
static FILE *popen(cmd, mode)
char *cmd, *mode
#endif
{
   char *s;

   TmpFile = tmpnam(NULL);
   if (!TmpFile) return NULL;
   s = (char *) malloc(strlen(cmd) + 3 + strlen(TmpFile) + 1);
   if (!s) return NULL;
   strcpy(s, cmd);
   strcat(s, " > ");
   strcat(s, TmpFile);
   system(s);
   free(s);
   return fopen(TmpFile, "r");
}

#ifdef HAVE_PROTOS
PRIVATE int pclose(FILE *fp)
#else
static int pclose(fp)
FILE *fp;
#endif
{
   unlink(TmpFile);
   return fclose(fp);
}

#endif

/***************************************************************/
/*                                                             */
/*  FTypeof                                                    */
/*                                                             */
/*  Implement the typeof() function.                           */
/*                                                             */
/***************************************************************/
#ifdef HAVE_PROTOS
PRIVATE int FTypeof(void)
#else
static int FTypeof()
#endif
{
   switch(ARG(0).type) {
      case INT_TYPE:  return RetStrVal("INT");
      case DATE_TYPE: return RetStrVal("DATE");
      case TIM_TYPE:  return RetStrVal("TIME");
      case STR_TYPE:  return RetStrVal("STRING");
      default:        return RetStrVal("ERR");
   }
}
